import os
import math
import time
import numpy as np

from .programParser import find_numbers, find_word
from .linalg import UR_2_Pose, Pose_2_TxyzRxyz, txyzRxyz_2_pose


class Postprocessor():
    """
    Generate program for Rozum Robotics robots like Pulse 75 or Pulse 90.

    Initial program is RoboDK generated program for Universal robot. That why we need to
    recalculate initial coordinates to the standart form (named - pose. 4x4 matrix) using UR_2_Pose function

    Should use Universal_Robots.py postprocessor with some changes.
    """
    PROG = ""
    PROG_EXT = 'py'
    TARGET_POSES = []
    TARGET_POSITIONS = []
    REF_FRAME = np.eye((4))
    TOOL_FRAME = np.zeros(6)
    SPEED_MS = 0.01

    def __init__(self, robothost: str):
        self.HOST = robothost
        self.progStart()


    def progSave(self, folder: str, progname: str, show_result: bool = False) -> None:
        """Save program in special folder"""
        progname = progname + '.' + self.PROG_EXT
        filesave = os.path.join(folder, progname)

        with open(filesave, 'w') as fid:
            fid.write(self.PROG)

        print('SAVED: %s\n' % filesave)
        # show result
        if show_result:
            # open file with default application
            os.startfile(filesave)


    def progStart(self) -> None:
        """Adds lines on the first place of your program"""
        self.addline("import time")
        self.addline('from pulseapi import *')
        self.addline(f"robot = RobotPulse('{self.HOST}')\n")


    def addline(self, newline: str) -> None:
        """Add a program line"""
        self.PROG = self.PROG + newline + '\n'


    def movel(self, line: str) -> None:
        """Recalculate UR coordinates to the standart form. """
        coordinates = np.array(find_numbers(line)[0:6])
        coordinates = UR_2_Pose(coordinates)
        coordinates = np.matmul(np.matmul(self.REF_FRAME, coordinates), np.linalg.inv(self.TOOL_FRAME))
        coordinates = np.matmul(self.REF_FRAME, coordinates)
        coordinates = Pose_2_TxyzRxyz(coordinates)
        XYZ = [round(value, 4) for value in coordinates[:3]]
        RXYZ = [round(value, 3) for value in coordinates[3:]]
        self.TARGET_POSITIONS.append(f'{XYZ}, {RXYZ}')


    def movej(self, line: str) -> None:
        coordinates = find_numbers(line)[0:6]
        coordinates = [round(math.degrees(coord), 2) for coord in coordinates]
        self.TARGET_POSES.append(f'{coordinates}')


    def runPositions(self, speed=None, velocity=None, acceleration=None) -> None:
        self.addline("target_position = [")
        for pos in self.TARGET_POSITIONS:
            self.addline(f"     position({pos}),")
        self.addline("]")
        self.TARGET_POSITIONS.clear()

        if isinstance(speed, (int, float)):
            self.addline(f'robot.run_positions(target_position, speed={speed}, motion_type=MT_LINEAR)')
            self.addline('robot.await_stop()')
        elif isinstance(velocity, (int, float)) and isinstance(acceleration, (int, float)) and speed == None:
            self.addline(f'robot.run_positions(target_position, velocity={velocity}, acceleration={acceleration}, motion_type=MT_LINEAR)')
            self.addline('robot.await_stop()')
        elif isinstance(velocity, (int, float)) and isinstance(acceleration, (int, float)) and isinstance(speed, (int, float)):
            self.addline(f'robot.run_positions(target_position, tcp_max_velocity={self.SPEED_MS}, motion_type=MT_LINEAR)')
            self.addline('robot.await_stop()')
        elif speed==None and velocity==None and acceleration==None:
            self.addline(f'robot.run_positions(target_position, tcp_max_velocity={self.SPEED_MS},  motion_type=MT_LINEAR)')
            self.addline('robot.await_stop()')
        else:
            raise AssertionError("Incorrect introduced parameters. Enter only speed parameter or velocity and acceleration")


    def runPoses(self, speed=None, velocity=None, acceleration=None):
        self.addline("target_poses = [")
        for pos in self.TARGET_POSES:
            self.addline(f"     pose({pos}),")
        self.addline("]")
        self.TARGET_POSES.clear()

        if isinstance(speed, (int, float)):
            self.addline(f'robot.run_poses(target_poses, speed={speed}, motion_type=MT_JOINT)')
            self.addline('robot.await_stop()')
        elif isinstance(velocity, (int, float)) and isinstance(acceleration, (int, float)) and speed == None:
            self.addline(f'robot.run_poses(target_poses, velocity={velocity}, acceleration={acceleration}, motion_type=MT_JOINT)')
            self.addline('robot.await_stop()')
        elif isinstance(velocity, (int, float)) and isinstance(acceleration, (int, float)) and isinstance(speed, (int, float)):
            self.addline(f'robot.run_poses(target_poses, tcp_max_velocity={self.SPEED_MS}, motion_type=MT_JOINT)')
            self.addline('robot.await_stop()')
        elif speed==None and velocity==None and acceleration==None:
            self.addline(f'robot.run_poses(target_poses, tcp_max_velocity={self.SPEED_MS}, motion_type=MT_JOINT)')
            self.addline('robot.await_stop()')
        else:
            raise AssertionError("Incorrect introduced parameters. Enter only speed parameter or velocity and acceleration")


    def set_tcp(self, line):
        """Set TCP frame
            :param line: line with tcp coordinates from UR program
        """
        self.TOOL_FRAME = np.array(find_numbers(line))
        self.TOOL_FRAME = UR_2_Pose(self.TOOL_FRAME)
        self.TOOL_FRAME = Pose_2_TxyzRxyz(self.TOOL_FRAME)
        x,y,z,r,p,w= self.TOOL_FRAME
        self.addline(f"robot.change_tool_info(tool_info(position([{x},{y},{z}], [{r},{p},{w}])))")
        ########################################
        self.TOOL_FRAME = txyzRxyz_2_pose(self.TOOL_FRAME)


    def set_ref_frame(self, positions):
        """Set reference frame position"""
        self.REF_FRAME = np.array(find_numbers(positions))
        self.REF_FRAME = UR_2_Pose(self.REF_FRAME)
        self.REF_FRAME = Pose_2_TxyzRxyz(self.REF_FRAME)
        x, y, z, r, p, w = self.REF_FRAME
        self.addline(f"robot.change_base(position([{self.REF_FRAME[:3]}], [{self.REF_FRAME[3:]}]))")


    def open_gripper(self):
        self.addline('robot.open_gripper()')


    def close_gripper(self):
        self.addline('robot.close_gripper()')


    def set_speed(self, line):
        """
        Set TCP speed, meters per second
        """
        self.SPEED_MS = find_numbers(line)[0]

        if float(self.SPEED_MS) > 2:
            raise AssertionError('Speed is too height. Set it lower then 2 ms')
        self.addline(f'speed_ms = {self.SPEED_MS}')

    def sleep(self, time):
        self.addline(f"time.sleep({time})")


import math
import numpy as np
from math import sqrt, atan2, pi, cos, sin

def txyzRxyz_2_pose(xyzrpw):
    """Returns the pose given the position (M) and Euler angles (rad) as an array [x,y,z,rx,ry,rz].
    The result is the same as calling: H = transl(x,y,z)*rotx(rx)*roty(ry)*rotz(rz)

    :param xyzrpw: [x,y,z,rx,ry,rz] in mm and radians
    :type xyzrpw: list of floats
    """
    [x,y,z,rx,ry,rz] = xyzrpw
    srx = math.sin(rx);
    crx = math.cos(rx);
    sry = math.sin(ry);
    cry = math.cos(ry);
    srz = math.sin(rz);
    crz = math.cos(rz);
    H = np.array([[ cry*crz, -cry*srz, sry, x],
                  [crx*srz + crz*srx*sry, crx*crz - srx*sry*srz, -cry*srx, y],
                  [srx*srz - crx*crz*sry, crz*srx + crx*sry*srz, crx*cry, z],
                  [0,0,0,1]])
    return H

def UR_2_Pose(xyzwpr):
    """Calculate the pose target given a p[x,y,z,u,v,w] cartesian target with rotation vector.
        This is the same format required by Universal Robot controllers.
    """
    x,y,z,w,p,r = xyzwpr
    wpr = [w,p,r]
    angle = norm(wpr)
    cosang = cos(0.5*angle)

    if angle == 0.0:
        q234 = [0.0,0.0,0.0]
    else:
        ratio = sin(0.5*angle)/angle
        q234 = mult3(wpr, ratio)

    q1234 = [cosang, q234[0], q234[1], q234[2]]
    pose = quaternion_2_pose(q1234)
    pose[0,3] = x
    pose[1,3] = y
    pose[2,3] = z
    return pose


def Pose_2_TxyzRxyz(H):
    """Retrieve the position (M) and Euler angles (rad) as an array [x,y,z,rx,ry,rz] given a pose.
    It returns the values that correspond to the following operation:
    H = transl(x,y,z)*rotx(rx)*roty(ry)*rotz(rz).
    """
    x = H[0,3]
    y = H[1,3]
    z = H[2,3]
    a = H[0,0]
    b = H[0,1]
    c = H[0,2]
    d = H[1,2]
    e = H[2,2]
    if c > (1.0 - 1e-6):
        ry1 = pi/2
        rx1 = 0
        rz1 = atan2(H[1,0],H[1,1])
    elif c < (-1.0 + 1e-6):
        ry1 = -pi/2
        rx1 = 0
        rz1 = atan2(H[1,0],H[1,1])
    else:
        sy = c
        cy1 = +sqrt(1-sy*sy)
        sx1 = -d/cy1
        cx1 = e/cy1
        sz1 = -b/cy1
        cz1 =a/cy1
        rx1 = atan2(sx1,cx1)
        ry1 = atan2(sy,cy1)
        rz1 = atan2(sz1,cz1)
    return [x, y, z, rx1, ry1, rz1]


def quaternion_2_pose(qin):
    """Returns the pose orientation matrix (4x4 matrix) given a quaternion orientation vector
    """
    qnorm = sqrt(qin[0]*qin[0]+qin[1]*qin[1]+qin[2]*qin[2]+qin[3]*qin[3])
    q = qin
    q[0] = q[0]/qnorm
    q[1] = q[1]/qnorm
    q[2] = q[2]/qnorm
    q[3] = q[3]/qnorm
    pose = np.array([[1 - 2*q[2]*q[2] - 2*q[3]*q[3], 2*q[1]*q[2] - 2*q[3]*q[0], 2*q[1]*q[3] + 2*q[2]*q[0], 0],
          [2*q[1]*q[2] + 2*q[3]*q[0], 1 - 2*q[1]*q[1] - 2*q[3]*q[3], 2*q[2]*q[3] - 2*q[1]*q[0], 0],
          [2*q[1]*q[3] - 2*q[2]*q[0], 2*q[2]*q[3] + 2*q[1]*q[0], 1 - 2*q[1]*q[1] - 2*q[2]*q[2], 0],
          [0 , 0 , 0 , 1]])
    return pose


def mult3(v,d):
    """Multiplies a 3D vector to a scalar"""
    return [v[0]*d, v[1]*d, v[2]*d]


def norm(p):
    """Returns the norm of a 3D vector"""
    return sqrt(p[0]*p[0] + p[1]*p[1] + p[2]*p[2])


def circle_radius(p0,p1,p2):
    a = norm(subs3(p0,p1))
    b = norm(subs3(p1,p2))
    c = norm(subs3(p2,p0))
    radius = a*b*c/sqrt(pow(a*a+b*b+c*c,2)-2*(pow(a,4)+pow(b,4)+pow(c,4)))
    return radius


def subs3(a,b):
    """Subtracts two 3D vectors c=a-b"""
    return [a[0]-b[0],a[1]-b[1],a[2]-b[2]]

def rot_x(rx):
    crx = math.cos(rx)
    srx = math.sin(rx)

    return np.array([[1,0,0,0], [0,crx,-srx,0], [0,srx,crx,0], [0,0,0,1]])

def rot_y(ry):
    crx = math.cos(ry)
    srx = math.sin(ry)

    return np.array([[crx,0,-srx,0], [0,1,0,0], [srx,0,crx,0], [0,0,0,1]])

def rot_z(rz):
    crx = math.cos(rz)
    srx = math.sin(rz)

    return np.array([[crx,-srx,0,0], [srx,crx,0,0], [0,0,1,0], [0,0,0,1]])

def transl(tx,ty=None,tz=None):
    """Returns a translation matrix (M)
    :param float tx: translation along the X axis
    :param float ty: translation along the Y axis
    :param float tz: translation along the Z axis
    """
    if ty is None:
        xx = tx[0]
        yy = tx[1]
        zz = tx[2]
    else:
        xx = tx
        yy = ty
        zz = tz
    return np.array([[1,0,0,xx],[0,1,0,yy],[0,0,1,zz],[0,0,0,1]])

def offset(target_position, refference_frame):
    """Calculates a relative target with respect to the reference frame coordinates.
    X,Y,Z are in mm, RX,RY,RZ are in degrees.
    """
    x, y, z, rx, ry, rz = refference_frame
    target_pose = xyzrpw_2_pose(target_position)

    # new_target = transl(x,y,z).dot(rot_x(rx*math.pi/180.0)).dot(rot_y(ry*math.pi/180.0)).dot(rot_z(rz*math.pi/180.0)).dot(target_pose)
    new_target = transl(x,y,z).dot(rot_x(rx)).dot(rot_y(ry)).dot(rot_z(rz)).dot(target_pose)
    new_target = pose_2_xyzrpw(new_target)

    return new_target


import re

def read_file(file_path):
    """Read initial program"""
    with open(file_path, 'r') as program:
        program = open(file_path, "r")
        program = str(program.read())
        return program.split('\n')

def find_numbers(data: str):
    """Finds all numbers in single line"""
    diction = re.findall('([-+]?(?:\d+(?:\.\d*)?|\.\d+))', data)
    diction = [float(number) for number in diction]
    return diction

def find_word(word: str, data: str) -> bool:
    """Find word in line"""
    if word in data:
        return True
    else: return False

def program_parser(program: str, postprocessor) -> None:
    """ """
    for i, line in enumerate(program):
        if find_word('set_ref', line):
            postprocessor.set_ref_frame(line)

        if find_word('set_tcp', line):
            postprocessor.set_tcp(line)

        if find_word('speed_ms', line) and find_word('=', line):
            postprocessor.set_speed(line)

        if find_word('movel([', line) and find_word('movel([', program[i+1]):
            postprocessor.movej(line)
        elif find_word('movel([', line):
            postprocessor.movej(line)
            postprocessor.runPoses()

        if find_word('movej', line) and find_word('movej', program[i+1]):
            postprocessor.movej(line)
        elif find_word('movej', line):
            postprocessor.movej(line)
            postprocessor.runPoses()

        if find_word('movel(p', line) and find_word('movel(p', program[i+1]):
            postprocessor.movel(line)
        elif find_word('movel(p', line):
            postprocessor.movel(line)
            postprocessor.runPositions()

        if find_word("Set air valve on", line):
            postprocessor.open_gripper()

        if find_word("Set air valve off", line):
            postprocessor.close_gripper()

        if find_word("Sleep", line):
            t = find_numbers(line)
            postprocessor.sleep(*t)


import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

VERSION = '0.1.8'
NAME = 'URToPulseConverter'
AUTHOR = 'Ostapets Vladislav'
AUTHOR_EMAIL = 'vlad12344444@gmail.com'
LONG_DESCRIPTION_CONTEXT_TYPE = 'text/markdown'
DESCRIPTION = "UR Robot script -> 'Rozum Robotics' Pulse Robots code converter "
PACKAGES = [
    'postprocessor',
    'processedPrograms',
    'RoboDKPrograms',
    'screenshots'
]
REQUIRED = [
    'numpy==1.18.0'
]

def read_long_description():
    try:
        path = os.path.join(here, "README.md")
        with open(path, "r") as ld:
            return ld.read()
    except FileNotFoundError:
        return DESCRIPTION

setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    description=DESCRIPTION,
    long_description=read_long_description(),
    long_description_content_type=LONG_DESCRIPTION_CONTEXT_TYPE,
    packages=PACKAGES,
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires='>=3.6',
    install_requires=REQUIRED,
    entry_points={
        "console_scripts": [
            "URToPulse=URScriptToPulseConverter.converter:main"
        ]
    }
)